import { computed, MaybeRefOrGetter, Ref, shallowRef, toValue, watchEffect } from 'vue'
import { makeRandomString } from '@/util'
import { Image, NewImage } from '@/types'

type AnyImage = string | Blob | ArrayBuffer

export function isNumber(v: unknown) {
  return typeof v === 'number' && !isNaN(v)
}

async function makeImageBlob(image: AnyImage | unknown) {
  if (typeof image === 'string') {
    return await fetch(image).then((res) => res.blob())
  } else if (image instanceof Blob) {
    return image
  } else if (image instanceof ArrayBuffer) {
    return new Blob([image])
  } else {
    throw new TypeError('Image must be a dataUrl, Blob or ArrayBuffer.')
  }
}

export function useImageBlob(initialValue: MaybeRefOrGetter<AnyImage | null>) {
  function setValue(value: AnyImage | null) {
    if (value === null) {
      imageBuffer.value = null
    } else {
      makeImageBlob(value)
        .then((buffer) => (imageBuffer.value = buffer))
        .catch((e) => {
          console.error('Unable to parse image as buffer.', value, e)
        })
    }
  }

  const imageBuffer = shallowRef<Blob | null>(null) as Ref<Blob | null>

  watchEffect(() => {
    setValue(toValue(initialValue))
  })

  return imageBuffer
}

export function useImageSource(image: MaybeRefOrGetter<Blob | null>) {
  return computed(() => {
    const value = toValue(image)
    if (!value) return undefined
    return URL.createObjectURL(value)
  })
}

export function useAspectRatio(
  dimensions: MaybeRefOrGetter<
    { width?: number | undefined; height?: number | undefined } | undefined
  >,
) {
  return computed(() => {
    const { width, height } = toValue(dimensions) ?? {}
    if (isNumber(width) && isNumber(height)) {
      return (width as number) / (height as number)
    } else {
      return undefined
    }
  })
}

export function newImage(
  file: File | Blob,
  originalImage?: Image | NewImage | undefined,
): NewImage {
  const attribtion = originalImage?.contentLicense

  return {
    url: URL.createObjectURL(file),
    file,
    altText: originalImage?.altText ?? '',
    ppoi: originalImage?.ppoi ?? '0.5x0.5',
    width: 0,
    height: 0,
    contentLicense: {
      licenseId: attribtion?.licenseId ?? null,
      credits: attribtion?.credits ?? '',
      isUseExplicitlyGrantedByAuthor: attribtion?.isUseExplicitlyGrantedByAuthor ?? false,
    },
  }
}

export function generateFileName(mimeType: string) {
  const name = 'image-' + makeRandomString(12)
  const ext = {
    'image/jpeg': 'jpg',
    'image/webp': 'webp',
    'image/avif': 'avif',
    'image/gif': 'gif',
    'image/png': 'png',
    'image/bmp': 'bmp',
  }[mimeType.toLowerCase()]
  if (!ext) {
    throw new Error('Invalid image mimetype.')
  }
  return `${name}.${ext}`
}
